from pathlib import Path
from typing import Union
import pydicom
import SimpleITK as itk
import numpy as np


def dcmread(
    path: Union[str, Path],
    read_header: bool = False,
    itk_handler_enabled: bool = True
):
    """ Read 2D image data from the DICOM file.

    Args:
        path (str or Path): path of the dicom file to be loaded.
        read_header (bool): whether to return the dicom header together
            with the image array.
        itk_handler_enabled (bool): whether to use SimpleITK to read the dicom
            if pydicom fails.

    Return:
        (ndarray): dicom image array.
        (pydicom.dataset.FileDataset): an instance of FileDataset
            that represents a parsed DICOM file.

    N.B.
        If itk_handler is enabled, segmentation fault (caused by SimpleITK
        reader) might happen which cannot be caught in python code.
    """
    ds = pydicom.dcmread(str(path))
    if itk_handler_enabled:
        try:
            img = ds.pixel_array
        except Exception:
            img_itk = itk.ReadImage(str(path))
            img = itk.GetArrayFromImage(img_itk)
            img = np.squeeze(img)
    else:
        img = ds.pixel_array

    if read_header:
        return img, ds
    else:
        return img


def dcminfo(path: Union[str, Path]):
    """ Read metadata (dicom tags) from DICOM file.

    Refer to pydicom.dcmread

    Args:
        path (str or Path): path of the dicom file.

    Return:
        (pydicom.dataset.FileDataset): an instance of FileDataset
        that represents a parsed DICOM file.
    """
    ds = pydicom.dcmread(str(path))
    return ds



from pathlib import Path
from typing import Union, Dict, List
import medvision as mv
import numpy as np
import pandas as pd


def load_det_dsmd(
    dsmd_path: Union[str, Path],
    class2label: Union[str, Dict[str, int]]
):
    """ load detection dataset metadata.

    Args:
        dsmd_path (str or Path): dataset metadata file path.
        class2label (str or dict): class-to-label file.

    Return:
        (OrderedDict): Loaded dsmd is a OrderedDict looks like
        {
            data/1.png: [
                bboxes (ndarray) of category 'cat' of shape (n, 4) or (n, 5),
                bboxes (ndarray) of category 'dog' of shape (n, 4) or (n, 5),
                ...
            ]
            data/2.png: [
                ...
            ]
            ...
        }
    """
    if isinstance(class2label, str):
        class2label = mv.load_c2l(class2label)
    assert min(class2label.values()) == 0, \
        "label should start from 0, but got %d" % min(class2label.values())
    num_classes = len(class2label)

    df = pd.read_csv(dsmd_path, header=None)
    assert df.shape[1] == 6 or df.shape[1] == 7, \
        "Incorrect dsmd file format %s" % dsmd_path

    data = {}
    for r in df.itertuples():
        filename = r[1]

        if filename not in data:
            empty_box = np.zeros((0, df.shape[1] - 2), dtype=np.float32)
            data[filename] = [empty_box] * num_classes

        if not pd.isnull(r[2]):
            box = r[2: -1]
            category_id = class2label[r[-1]]
            data[filename][category_id] = np.append(
                data[filename][category_id], [box], axis=0
            )

    return mv.make_dsmd(data)


def save_det_dsmd(
    dsmd_path: Union[str, Path],
    data: Dict[str, List[np.ndarray]],
    class2label: Union[str, Dict[str, int]],
    auto_mkdirs: bool = True
):
    """ Save dataset metadata to specified file.

    Args:
        dsmd_path (str or Path): file path to save dataset metadata.
        data (dict): dsmd to be serialized, refer to 'load_dsmd'.
        class2label (str or dict): class-to-label file or class2label dict.
        auto_mkdirs (bool): If the parent folder of `file_path` does not
            exist, whether to create it automatically.
    """
    if auto_mkdirs:
        mv.mkdirs(mv.parentdir(dsmd_path))

    # get label->class mapping
    if isinstance(class2label, str):
        class2label = mv.load_c2l(class2label)
    label2class = {value: key for key, value in class2label.items()}

    # write dataset metadata loop
    dsmd = mv.make_dsmd(data)
    data_list = []
    for filename, instance in dsmd.items():
        record = None
        for category_id, boxes in enumerate(instance):
            for box in boxes:
                record = [filename, *box.tolist(), label2class[category_id]]
                data_list.append(record)
        if record is None:
            data_list.append([filename])

    df = pd.DataFrame(data_list)
    df.to_csv(str(dsmd_path), header=False, index=False)


def merge_det_dsmds(ref_dsmd, *dsmds):
    """ Merge dsmds into one dsmd.

    N.B. Overlapping boxes (even boxes with the same coordinates) are all kept.
    """
    filenames = set(ref_dsmd.keys())
    assert len(filenames) > 0
    for dsmd in dsmds:
        assert set(dsmd.keys()) == filenames, "dsmds not match"

    ref_filename = filenames.pop()
    num_categories = len(ref_dsmd[ref_filename])
    assert num_categories != 0
    filenames.add(ref_filename)

    res = {}
    for filename in filenames:
        res_instance = []
        for category_id in range(num_categories):
            instances = [dsmd[filename][category_id] for dsmd in dsmds] \
                        + [ref_dsmd[filename][category_id]]
            res_instance.append(np.vstack(instances))
        res[filename] = res_instance

    return res



import ftplib
import logging
import os
import socket
import sys


class FTP:
    ftp = ftplib.FTP()

    def __init__(self, host, port=21):
        self.host = host
        self.ftp.connect(host, port)

    def login(self, user, passwd):
        try:
            self.ftp.login(user, passwd)
            logging.info('Connection established: [%s]', self.host)
        except (socket.error, socket.gaierror):
            logging.error('FTP login failed, please check correctness of '
                          'host machine ip, user name and password.')
            sys.exit(0)

    def logout(self):
        logging.info('Disconnected from server: [%s]', self.host)
        self.ftp.quit()

    def download_file(self, local_file, remote_file):
        file_handler = open(local_file, 'wb')
        logging.info(file_handler)
        self.ftp.retrbinary('RETR ' + remote_file, file_handler.write)
        file_handler.close()
        logging.info('[%s] File transfer successful.', remote_file)
        return True

    def is_ftp_dir(self, file_name):
        try:
            self.ftp.cwd(file_name)
            self.ftp.cwd('..')
            return True
        except ftplib.error_perm:
            return False

    def download_tree(self, local_dir, remote_dir):
        logging.info('remote_dir, %s', remote_dir)
        if not os.path.exists(local_dir):
            os.makedirs(local_dir)
        self.ftp.cwd(remote_dir)
        remote_names = self.ftp.nlst()
        logging.info('remote_names, %s', remote_names)
        for file in remote_names:
            local = os.path.join(local_dir, file)
            logging.info(self.ftp.nlst(file))
            if self.is_ftp_dir(file):
                if not os.path.exists(local):
                    os.makedirs(local)
                self.download_tree(local, file)
            else:
                self.download_file(local, file)

        self.ftp.cwd('..')
        return


if __name__ == '__main__':
    logging.getLogger().setLevel(logging.INFO)

    local_dir = 'test_data'
    remote_dir = '/release-dependencies/'

    HOST_216 = '192.168.1.216'
    USER_216 = 'reader'
    PASSWD_216 = 'reader'

    ftp216 = FTP(HOST_216)
    ftp216.login(user=USER_216, passwd=PASSWD_216)
    ftp216.download_tree(local_dir, remote_dir)
    ftp216.logout()



# flake8: noqa

from .colorspace import (rgb2gray, gray2rgb)
from .geometry import (vflip, hflip, rot90, rotate, resize, rescale,
                       crop, center_crop, pad_to_square)
from .normalize import (
    normalize_grayscale, normalize_rgb, denormalize_rgb, imadjust_grayscale
)

__all__ = [k for k in globals().keys() if not k.startswith("_")]



import json
import cv2
import medvision as mv
from tqdm import tqdm


def _gen_rws_shape(label, contour):
    line_color = (0, 255, 0, 128)
    fill_color = (255, 0, 0, 128)
    flags = {}
    shape_type = 'polygon'

    return dict(
        label=label,
        line_color=line_color,
        fill_color=fill_color,
        points=[(float(p[0][0]), float(p[0][1])) for p in contour],
        shape_type=shape_type,
        flags=flags
    )


def mask2rws(mask_path, rws_path, dcm_path=None):
    if isinstance(mask_path, str):
        mask = mv.imread(mask_path, mv.ImreadMode.GRAY)
        dcm_path = mv.basename(mask_path).replace('.png', '.dcm')
    else:
        mask = mask_path
        assert dcm_path is not None

    if mask.max() == 0:
        print('skip mask with no annotation regions: {}'.format(mask_path))
        return

    contours, _ = cv2.findContours(
        mask,
        cv2.RETR_TREE,
        cv2.CHAIN_APPROX_SIMPLE
    )

    shapes = []
    for contour in contours:
        shape = _gen_rws_shape('auto', contour)
        shapes.append(shape)

    data = dict(
        version='0.1.0',
        flags={},
        shapes=shapes,
        lineColor=None,
        fillColor=None,
        imagePath=dcm_path,
        imageData=None,
        imageHeight=mask.shape[0],
        imageWidth=mask.shape[1],
    )

    with open(rws_path, 'w') as f:
        json.dump(data, f, ensure_ascii=False, indent=2)


def batch_mask2rws(mask_dir, rws_dir, **kwargs):
    """ Convert mask format annotation to rws format.

    Args:
        mask_dir (str): mask files directory.
        rws_dir (str): rws annotation files directory.

    N.B. dicom file title should be exactly the same with mask file title.
    e.g. 123.dcm, 123.png
    """
    mv.mkdirs(rws_dir)
    mask_filenames = mv.listdir(mask_dir)

    file_titles = [mv.splitext(fn)[0] for fn in mask_filenames]

    for file_title in tqdm(file_titles):
        mask_path = mv.joinpath(mask_dir, file_title + '.png')
        rws_path = mv.joinpath(rws_dir, file_title + '.json')
        mask2rws(mask_path, rws_path, **kwargs)


if __name__ == '__main__':
    # TODO: move to unittest
    mask_dir = '/home/huiying/.rws/masks'
    rws_dir = '/home/huiying/.rws/rws_verification'

    batch_mask2rws(mask_dir, rws_dir)



# flake8: noqa

from .fileutil import (isdir, isfile, listdir,
                       joinpath, basename, abspath, splitext,
                       parentdir, filetitle, change_suffix,
                       cp, rm, cptree, rmtree, move, symlink,
                       mkdirs, empty_dir,
                       non_overwrite_cp, copyfiles,
                       GlobMode, glob,
                       find_duplicated_files)
from .multiprocessutil import tqdm_imap_unordered
from .timer import Timer
from .typeutil import isarrayinstance


__all__ = [k for k in globals().keys() if not k.startswith("_")]



# flake8: noqa

from .dr_io import DrReadMode, dcmread_dr
from .fileutil import isdicom, isdicomdir
from .io import dcmread, dcminfo


__all__ = [k for k in globals().keys() if not k.startswith("_")]



import multiprocessing
from tqdm import tqdm


def tqdm_imap_unordered(func, args, n_processes=None):
    """ Parallel processing with a progress bar.

    Run a function multiple times (with different inputs) utilizing parallel
    computing. During the parallel processing a progress bar is shown to
    check the processing progress.

    N.B. The ordering of results is arbitrary.

    Args:
        func (function obj): the function to be called.
        args (list): the function arguments.
        n_processes (int): number of processes to be utilized.

    Return:
        (list): the results for different inputs.
    """
    p = multiprocessing.Pool(n_processes)

    result = []
    for res in tqdm(p.imap_unordered(func, args), total=len(args)):
        result.append(res)

    p.close()
    p.join()

    return result


# Multiprocessing cannot be tested in pytest.
# We can test it here instead.
if __name__ == '__main__':
    def add(a, b, c):
        return a + b + c

    first = [str(i) for i in range(10)]
    second = '_m'
    third = '.png'

    # single processor version
    result_sp = [add(i, second, third) for i in first]

    # multi-processors version 1
    def add_wrapper(xyz):
        return add(*xyz)
    in_para_wrapper = [(i, second, third) for i in first]
    result_mp = tqdm_imap_unordered(add_wrapper, in_para_wrapper)
    print(result_mp)

    # multi-processors version 2
    from functools import partial
    partial_func = partial(add, b=second, c=third)
    result_mp = tqdm_imap_unordered(partial_func, first)
    print(result_mp)



import json
import medvision as mv


def get_rws_datainfo_path(dcm_path):
    return mv.change_suffix(dcm_path, ".datainfo")


def get_rws_annot_path(dcm_path):
    return mv.change_suffix(dcm_path, ".json")


def get_rws_flag_path(dcm_path):
    return mv.change_suffix(dcm_path, ".flag")


def get_rws_text_path(dcm_path):
    return mv.change_suffix(dcm_path, ".text")


def load_rws_contour(filepath):
    """ Load rws file (in general contour format).

    Args:
        filepath (str): rws file path.

    Returns:
        rws object (dict), e.g.
        {
            'height': 256,
            'width': 245,
            'image_path': 'example.dcm',
            'shapes':
            [
                {
                    'category': 'cat',
                    'points': [[12, 12], [34, 34], ...]
                },
                {
                    'category': 'dog',
                    'points': [[34, 34], [56, 56], ...]
                },
            ]
        }
    """
    with open(filepath, 'r', encoding='utf-8') as f:
        data = json.load(f)

    # relative path from label file to relative path from cwd
    image_path = mv.joinpath(mv.parentdir(filepath), data['imagePath'])
    height = data.get('imageHeight')
    width = data.get('imageWidth')
    shapes = [
        {'category': s['label'], 'points': s['points']}
        for s in data['shapes']
    ]

    return {
        'height': height,
        'width': width,
        'image_path': image_path,
        'shapes': shapes
    }


def load_rws_bbox(filepath):
    """ Load rws file (in bounding box format).

    Args:
        filepath (str): rws file path.

    Returns:
        rws object (dict), e.g.
        {
            'height': 256,
            'width': 245,
            'image_path': 'example.dcm',
            'shapes':
            [
                {
                    'category': 'cat',
                    'bbox': [12, 12, 34, 34]
                },
                {
                    'category': 'dog',
                    'bbox': [34, 34, 56, 56]
                },
            ]
        }

    N.B. this function only support rectangle annotation.
    """
    rws = load_rws_contour(filepath)

    shapes = []
    for instance in rws['shapes']:
        category, points = instance['category'], instance['points']
        assert len(points) == 2, 'only support rectangle annotation'
        xmin = min(points[0][0], points[1][0])
        ymin = min(points[0][1], points[1][1])
        xmax = max(points[0][0], points[1][0])
        ymax = max(points[0][1], points[1][1])
        shape = {
            'category': category,
            'bbox': [xmin, ymin, xmax, ymax]
        }
        shapes.append(shape)

    rws['shapes'] = shapes

    return rws


def _gen_rws_shape_bbox(bbox, label):
    flags = {}
    shape_type = 'rectangle'
    line_color = (0, 255, 0, 128)
    fill_color = (255, 0, 0, 128)
    return dict(
        label=label,
        line_color=line_color,
        fill_color=fill_color,
        points=[[float(bbox[0]), float(bbox[1])],
                [float(bbox[2]), float(bbox[3])]],
        shape_type=shape_type,
        flags=flags
    )


def save_rws_bbox(
    filepath,
    shapes,
    image_shape,
    suffix=".json",
    score_thresh=0,
    fixed_label="auto"
):
    """ Save bounding boxes into rws file.

    Args:
        filepath (str): rws file path.
        shapes (list or numpy.ndarray): "shape" field in rws file or
            raw bounding boxes (in numpy.ndarray format or list format,
            in this case, only single-class label is supported).
        image_shape (tuple): shape of corresponding image, in (h, w) format.
        suffix (str): rws file suffix, e.g. ".json", ".json_EA", ".json_A1".
        score_thresh (float): probability threshold of bounding boxes.
            If probability score of a box is larger than the threshold,
            the box will be saved into rws file, otherwise not.
        fixed_label (str): bounding box label (for 1 class object detection).
            if "shapes" is a numpy.ndarray, or box list, this label is used.
    """
    assert len(shapes) != 0, "no bounding box for [%s]" % filepath
    assert suffix.startswith(".json"), "unsupported suffix %s" % suffix

    # if not default rws suffix, change suffix
    if suffix != ".json":
        filepath = mv.change_suffix(filepath, suffix)

    # for raw bounding boxes, convert to rws "shape" field format first.
    bboxes = shapes
    shapes = []
    for bbox in bboxes:
        if isinstance(bbox, dict):  # rws "shape" field format
            label, bbox = bbox['category'], bbox['bbox']
        elif len(bbox) == 5 and bbox[-1] < score_thresh:
            continue
        else:  # ndarray or box list
            label, bbox = fixed_label, bbox[:4]
        shape = _gen_rws_shape_bbox(bbox, label)
        shapes.append(shape)

    # make rws file content
    data = dict(
        version='0.1.0',
        flags={},
        shapes=shapes,
        lineColor=None,
        fillColor=None,
        imagePath=mv.filetitle(filepath) + ".dcm",
        imageData=None,
        imageHeight=image_shape[0],
        imageWidth=image_shape[1],
    )

    # save rws file
    with open(filepath, 'w') as f:
        json.dump(data, f, ensure_ascii=False, indent=2)



# flake8: noqa

from .mask2rws import batch_mask2rws, mask2rws
from .rws2mask import batch_rws2mask, rws2mask
from .rws2dsmd import rws2dsmd_bbox, dsmd2rws_bbox
from .rws import (
    get_rws_datainfo_path,
    get_rws_annot_path,
    get_rws_flag_path,
    get_rws_text_path,
    load_rws_contour,
    load_rws_bbox,
    save_rws_bbox
)


__all__ = [k for k in globals().keys() if not k.startswith("_")]



from typing import Sequence, Optional
import matplotlib.pyplot as plt
from matplotlib.ticker import FixedFormatter
import numpy as np
import sklearn


def draw_froc_curve(average_fps,
                    sensitivity,
                    save_path=None,
                    bLogPlot=True,
                    **kwargs):
    """ Plot the FROC curve.

    Args:
        average_fps (list): A list containing the average number of false
        positives per image for different thresholds.
        sensitivity (list):  A list containing overall sensitivity for
        different thresholds.
        save_path (str): path to save the froc curve drawing.
    """
    average_fps = np.append(average_fps, 64)
    sensitivity = np.append(sensitivity, sensitivity[-1])
    plt.figure()
    plt.plot(average_fps, sensitivity, **kwargs)
    plt.xlabel('Average number of false positives per scan')
    plt.ylabel('Sensitivity')
    plt.title('FROC Curve')
    plt.xlim(0.125, 64)
    plt.ylim([0, 1])
    plt.legend(loc='lower right')

    ax = plt.gca()
    xaxis = [0.125, 0.25, 0.5, 1, 2, 4, 8, 16, 32, 64]
    if bLogPlot:
        plt.xscale('log', basex=2)
        ax.xaxis.set_major_formatter(FixedFormatter(xaxis))

    # set your ticks manually
    ax.xaxis.set_ticks(xaxis)
    ax.yaxis.set_ticks(np.arange(0, 1.1, 0.1))
    plt.grid(b=True, which='both')
    plt.tight_layout()

    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()


def draw_roc_curve(
    fpr: Sequence[float],
    tpr: Sequence[float],
    save_path: Optional[str] = None
):
    roc_auc = sklearn.metrics.auc(fpr, tpr)
    plt.figure()
    plt.plot(fpr, tpr, color='darkorange', lw=2,
             label='ROC curve (auc = %0.4f)' % roc_auc)
    plt.plot([0, 1], [0, 1], color='navy', lw=2, linestyle='--')
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.xlabel('FPR')
    plt.ylabel('TPR')
    plt.xticks()
    plt.yticks()
    plt.title('ROC Curve')
    plt.legend(loc="lower right")
    plt.grid(b=True, which='both')
    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()


def draw_pr_curve(
    precision: Sequence[float],
    recall: Sequence[float],
    save_path: Optional[str] = None
):
    pr_auc = sklearn.metrics.auc(precision, recall)
    plt.figure()
    plt.plot(recall, precision, color='darkorange', lw=2,
             label='PR curve (auc = %0.4f)' % pr_auc)
    plt.xlim([-0.01, 1.01])
    plt.ylim([-0.01, 1.01])
    plt.xlabel('Recall')
    plt.ylabel('Precision')
    plt.xticks()
    plt.yticks()
    plt.title('PR Curve')
    plt.legend(loc="lower right")
    plt.grid(b=True, which='both')
    if save_path is not None:
        plt.savefig(save_path)
    else:
        plt.show()



# flake8: noqa

from .draw_curve import draw_froc_curve, draw_roc_curve, draw_pr_curve
from .image import Color, imshow, imshow_bboxes

__all__ = [k for k in globals().keys() if not k.startswith("_")]



from enum import Enum, unique
from pathlib import Path
from typing import Union
import cv2
import medvision as mv
import numpy as np


@unique
class ImreadMode(Enum):
    RGB = cv2.IMREAD_COLOR
    GRAY = cv2.IMREAD_GRAYSCALE
    UNCHANGED = cv2.IMREAD_UNCHANGED


def imread(
    file_path: Union[str, Path],
    flag: ImreadMode = ImreadMode.RGB
):
    """ Read an image.

    Args:
        file_path (str or Path): image file path.
        flag (ImreadMode): flags specifying the color type of a loaded image,
              refer to ImreadMode for more details.
    Returns:
        (numpy.ndarray): loaded image array.

    Note:
        The format of the loaded image array is RGB.
    """
    img = cv2.imread(str(file_path), flag.value)

    if img.ndim == 3:
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)

    return img


def imwrite(
    file_path: Union[str, Path],
    img: np.ndarray,
    auto_mkdirs: bool = True
):
    """ Save image to specified file.

    Args:
        file_path (str or Path): specified file path to save to.
        img (numpy.ndarray): image array to be written.
        auto_mkdirs (bool): If the parent folder of `file_path` does not exist,
            whether to create it automatically.

    Returns:
        (bool): returns whether the image is saved successfully.

    Note:
        If the given image is a color image. It should be in RGB format.
    """
    if auto_mkdirs:
        mv.mkdirs(mv.parentdir(file_path))

    if img.ndim == 3:
        img = cv2.cvtColor(img, cv2.COLOR_RGB2BGR)

    return cv2.imwrite(str(file_path), img)



from enum import Enum, unique
from pathlib import Path
from typing import Union
import pydicom
import medvision as mv


@unique
class DrReadMode(Enum):
    MONOCHROME1 = 1
    MONOCHROME2 = 2
    UNCHANGED = None


def _invert_if_needed(img, mode, mono):
    # Convert 'Photometric Interpretation' if needed
    mono = 1 if mono.upper().find('MONOCHROME1') != -1 else 2
    if mode.value != mono and mode.value is not DrReadMode.UNCHANGED.value:
        return img.max() - img + img.min(), True
    else:
        return img, False


def dcmread_dr(
    path: Union[str, Path],
    mode: DrReadMode = DrReadMode.MONOCHROME2,
    read_header: bool = False,
    itk_handler_enabled: bool = True
):
    """ Read 2D digital radiography image data from the DICOM file.

    Args:
        path (str or Path): path of the dicom file.
        mode ('DrReadMode'): read mode, refer to 'DrReadMode'.
        read_header (bool): whether to return the dicom header together
            with the image array.
        itk_handler_enabled (bool): whether to use SimpleITK to read the
            dicom if pydicom fails. N.B. SimpleITK reader may cause
            segmentation fault which cannot be caught in python code.
    Returns:
        (ndarray): dicom image array.
        (dict, optional): dicom metadata.

    Note:
        MONOCHROME1 indicates that the greyscale ranges from bright to dark
        with ascending pixel values, whereas MONOCHROME2 ranges from dark
        to bright with ascending pixel values.
    """
    img, ds = mv.dcmread(
        path,
        read_header=True,
        itk_handler_enabled=itk_handler_enabled
    )

    if img.ndim == 3:
        img = img[:, :, 0]

    # fetch monochrome value
    assert 'PhotometricInterpretation' in ds
    mono = ds.PhotometricInterpretation

    img, is_inverted = _invert_if_needed(img, mode, mono)

    if read_header:
        if 'WindowCenter' in ds and 'WindowWidth' in ds:
            if isinstance(ds.WindowCenter, pydicom.multival.MultiValue):
                ds.WindowCenter = ds.WindowCenter[0]
                ds.WindowWidth = ds.WindowWidth[0]
            ds.WindowCenter = float(ds.WindowCenter)
            ds.WindowWidth = float(ds.WindowWidth)
            if is_inverted:
                ds.WindowCenter = (
                    float(img.max()) + float(img.min()) - ds.WindowCenter
                )
        if is_inverted:
            ds.PhotometricInterpretation = mode.name
        return img, ds
    else:
        return img



import numpy as np
import medvision as mv


def rws2dsmd_bbox(
    filepaths,
    class2label,
    suffix='.json',
    num_classes=None
):
    """ Convert rws bbox annotation to dsmd.

    Args:
        filepaths (str or list): file paths of rws annotation files or
            directory containing rws annotation files.
        class2label (str, dict or callable): class-to-label file,
            class2label dict or a callable.
        suffix (str): suffix of rws file. e.g. ".json", ".json_A1"
        num_classes (int or None): if class2label is a callable, num_classes
            should be explicitly specified.
    N.B.
        Key values of dsmd are file titles.
    """
    if callable(class2label):
        assert num_classes is not None,\
            'must specify num_classes if class2label is a callable'
    else:
        if isinstance(class2label, str):
            class2label = mv.load_c2l(class2label)
        num_classes = len(class2label)

    if isinstance(filepaths, str):
        filepaths = mv.glob(filepaths, '*' + suffix)

    dsmd = {}
    for filepath in filepaths:
        key = mv.filetitle(filepath)
        dsmd[key] = [np.zeros((0, 4), dtype=np.float32)] * num_classes
        # to handle cases without annot when getting annot path from image path
        if not mv.isfile(filepath):
            continue
        shapes = mv.load_rws_bbox(filepath)['shapes']
        for instance in shapes:
            try:
                label = class2label[instance['category']]
            except TypeError:
                label = class2label(instance['category'])
            dsmd[key][label] = np.append(
                dsmd[key][label], [instance['bbox']], axis=0
            )
    return mv.make_dsmd(dsmd)


def dsmd2rws_bbox(
    rws_dir,
    dsmd,
    class2label,
    suffix='.json_A1',
    score_thresh=0.0,
    dcm_dir=None
):
    """ Convert dsmd to rws bbox annotations.
    Args:
        rws_dir (str): directory to store converted rws file.
        dsmd (str or dsmd): file path of dsmd file or dsmd data.
        class2label (str or dict): class-to-label file or class2label dict.
        suffix (str): suffix of output rws file, For example, ".json_EA".
        score_thresh (float): threshold of confidence.
        dcm_dir (None or str): directory containing corresponding dicom files,
            if None, it is the same with rws_dir.
    N.B.
        1. Key values of dsmd are file titles.
        2. dicom file titles must be consistent with dsmd keys.
    """
    if isinstance(class2label, str):
        class2label = mv.load_c2l(class2label)
    label2class = {v: k for k, v in class2label.items()}

    if isinstance(dsmd, str):
        dsmd = mv.load_dsmd(dsmd, class2label, mode='det')

    if dcm_dir is None:
        dcm_dir = rws_dir
    dcm_titles = [mv.filetitle(path) for path in mv.glob(dcm_dir, "*.dcm")]
    assert set(dsmd.keys()) == set(dcm_titles)

    for key in dsmd.keys():
        shapes = []
        for i, bboxes in enumerate(dsmd[key]):
            for bbox in bboxes:
                if len(bbox) == 5 and bbox[-1] < score_thresh:
                    continue
                label = label2class[i] if len(bbox) == 4 else \
                    label2class[i] + "_{:.3f}".format(bbox[-1])
                shapes.append(
                    {'category': label, 'bbox': bbox[:4].tolist()}
                )
        if len(shapes) == 0:
            continue

        rws_path = mv.joinpath(rws_dir, key + suffix)
        ds = mv.dcminfo(mv.joinpath(dcm_dir, key + '.dcm'))
        mv.save_rws_bbox(rws_path, shapes, (ds.Rows, ds.Columns))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
from setuptools import find_packages, setup
from setuptools.extension import Extension

from Cython.Build import cythonize
import numpy


NAME = 'medvision'
DESCRIPTION = 'A python library for medical image computer vision.'
URL = 'https://github.com/kyle0x54/medvision'
EMAIL = 'kyle0x54@163.com'
AUTHOR = 'kyle0x54'
REQUIRES_PYTHON = '>=3.6.0'
VERSION = None


REQUIRED = [
    'cython',
    'matplotlib',
    'natsort',
    'numpy',
    'opencv-python',
    'pandas',
    'pillow',
    'pydicom',
    'scikit-learn',
    'SimpleITK',
    'tqdm',
]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with open(os.path.join(here, 'README.md')) as f:
        long_description = '\n' + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, '__version__.py')) as f:
        exec(f.read(), about)
else:
    about['__version__'] = VERSION

extensions = [
    Extension(
        'medvision.evaluation.compute_overlap',
        ['medvision/evaluation/compute_overlap.pyx'],
        include_dirs=[numpy.get_include()]),
    ]


setup(
    name=NAME,
    version=about['__version__'],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=find_packages(exclude=('tests',)),
    install_requires=REQUIRED,
    include_package_data=True,
    license='APACHE',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    zip_safe=False,
    ext_modules=cythonize(extensions),
)



from pathlib import Path
from typing import Union
import medvision as mv


def isdicom(path: Union[str, Path]):
    """ Judge whether a given file is a valid dicom.

    Args:
        path(str or Path): given file path.

    Returns:
        (bool): True if given path is a valid dicom, otherwise False.
    """
    if not mv.isfile(path):
        return False

    # read preamble and magic code
    with open(path, 'rb') as f:
        header = f.read(132)

    if not header:
        return False

    # magic code of a dicom file should be 'DICM'
    magic_code = header[128:132]
    if magic_code != b'DICM':
        return False
    else:
        return True


def isdicomdir(path: Union[str, Path]):
    """ Judge whether a given directory is a valid dicom directory.

    If given directory only contains dicoms (at least one dicom file),
    it is a dicom directory. Otherwise, it is not a dicom directory.

    Args:
        path(str): given directory path.

    Returns:
        (bool): True if given directory path is a dicom directory,
                otherwise False.
    """
    if not mv.isdir(path):
        return False

    for file_name in mv.listdir(path):
        file_path = mv.joinpath(path, file_name)
        if not isdicom(file_path):
            return False
    else:
        return True



import math
import PIL.Image
import PIL.ImageDraw
import medvision as mv
import numpy as np
from tqdm import tqdm

from .rws import load_rws_contour


def shape_to_mask(img_shape, points, shape_type=None,
                  line_width=10, point_size=5):
    mask = np.zeros(img_shape[:2], dtype=np.uint8)
    mask = PIL.Image.fromarray(mask)
    draw = PIL.ImageDraw.Draw(mask)
    xy = [tuple(point) for point in points]
    if shape_type == 'circle':
        assert len(xy) == 2, 'Shape of shape_type=circle must have 2 points'
        (cx, cy), (px, py) = xy
        d = math.sqrt((cx - px) ** 2 + (cy - py) ** 2)
        draw.ellipse([cx - d, cy - d, cx + d, cy + d], outline=1, fill=1)
    elif shape_type == 'rectangle':
        assert len(xy) == 2, 'Shape of shape_type=rectangle must have 2 points'
        draw.rectangle(xy, outline=1, fill=1)
    elif shape_type == 'line':
        assert len(xy) == 2, 'Shape of shape_type=line must have 2 points'
        draw.line(xy=xy, fill=1, width=line_width)
    elif shape_type == 'linestrip':
        draw.line(xy=xy, fill=1, width=line_width)
    elif shape_type == 'point':
        assert len(xy) == 1, 'Shape of shape_type=point must have 1 points'
        cx, cy = xy[0]
        r = point_size
        draw.ellipse([cx - r, cy - r, cx + r, cy + r], outline=1, fill=1)
    else:
        if len(xy) > 2:
            draw.polygon(xy=xy, outline=1, fill=1)
        else:
            pass
    mask = np.array(mask, dtype=bool)
    return mask


def rws2mask(rws_path, mask_path):
    rws_contour = load_rws_contour(rws_path)

    img_shape = (rws_contour['height'], rws_contour['width'])
    mask = np.zeros(img_shape[:2], dtype=bool)
    shapes = rws_contour['shapes']
    for shape in shapes:
        m = shape_to_mask(img_shape, shape['points'])
        mask = np.bitwise_xor(mask, m)
    mask = mask.astype(np.uint8) * 255

    mv.imwrite(mask_path, mask)


def batch_rws2mask(rws_dir, mask_dir, **kwargs):
    """ Convert rws format annotation to mask format.

    Args:
        rws_dir (str): rws annotation files directory.
        mask_dir (str): mask files directory.

    N.B. dicom file title should be exactly the same with mask file title.
    e.g. 123.dcm, 123.png
    """
    mv.mkdirs(mask_dir)
    rws_filenames = mv.listdir(rws_dir)

    file_titles = [mv.splitext(fn)[0] for fn in rws_filenames]

    for file_title in tqdm(file_titles):
        rws_path = mv.joinpath(rws_dir, file_title + '.json')
        mask_path = mv.joinpath(mask_dir, file_title + '.png')
        rws2mask(rws_path, mask_path, **kwargs)


if __name__ == '__main__':
    # TODO: move to unittest
    rws_dir = '/home/huiying/.rws/rws'
    mask_dir = '/home/huiying/.rws/masks'

    batch_rws2mask(rws_dir, mask_dir)



# flake8: noqa

from .transforms import (rgb2gray, gray2rgb,
                         vflip, hflip, rot90, rotate, resize, rescale, crop,
                         center_crop, pad_to_square,
                         normalize_grayscale, normalize_rgb, denormalize_rgb,
                         imadjust_grayscale)

from .io import ImreadMode, imread, imwrite

__all__ = [k for k in globals().keys() if not k.startswith("_")]



# flake8: noqa

from .dsmd import load_dsmd, save_dsmd, load_c2l
from .classification import (gen_cls_dsmd_file_from_datafolder,
                             gen_cls_ds_from_datafolder)
from .util import make_dsmd, split_dsmd_file, match_dsmds, update_dsmd_keys
from .detection import load_det_dsmd, save_det_dsmd, merge_det_dsmds
from .segmentation import load_seg_dsmd, save_seg_dsmd


__all__ = [k for k in globals().keys() if not k.startswith("_")]
